/********************************************************* begin of preamble
**
** Copyright (C) 2003-2012 Software- und Organisations-Service GmbH. 
** All rights reserved.
**
** This file may be used under the terms of either the 
**
**   GNU General Public License version 2.0 (GPL)
**
**   as published by the Free Software Foundation
**   http://www.gnu.org/licenses/gpl-2.0.txt and appearing in the file
**   LICENSE.GPL included in the packaging of this file. 
**
** or the
**  
**   Agreement for Purchase and Licensing
**
**   as offered by Software- und Organisations-Service GmbH
**   in the respective terms of supply that ship with this file.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
** IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
** THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
** BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
** CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
** POSSIBILITY OF SUCH DAMAGE.
********************************************************** end of preamble*/
package com.sos.testframework.h2;

import com.google.common.io.Files;
import com.google.common.io.Resources;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.nio.charset.Charset;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

/**
 * A general abstract class for a H2 connection.
 *
 * @uthor Stefan Schädlich
 * @version 29.05.13 12:57
 */
public abstract class H2Connection {

    private static final URL HIBERNATE_CFG_TEMPLATE	= Resources.getResource("com/sos/testframework/h2/hibernate-h2.cfg.xml.template");

    private final static Logger logger = LoggerFactory.getLogger(H2Connection.class);

    private final String connectionPrefix;
    private final List<File> sqlFiles;
    private final List<String> classNames;

    private Connection connection = null;
    private String initString = null;
    private String connectionString = null;

    private File temporaryWorkingDirectory = null;
    private File temporaryConfigFile = null;

    protected H2Connection(H2ConnectionType type) {
        this.connectionPrefix = type.getPrefix();
        this.sqlFiles = new ArrayList<File>();
        this.classNames = new ArrayList<String>();
    }

    protected H2Connection(H2ConnectionType type, ResourceList fileList) {
        this.connectionPrefix = type.getPrefix();
        this.sqlFiles = fileList.getFilelist();
        this.classNames = fileList.getClasslist();
    }

    private String getClassList() {
        StringBuffer result = new StringBuffer();
        for(String className : classNames) {
            result.append("<mapping class=\"");
            result.append(className);
            result.append("\"/>");
            result.append("\n");
        }
        return result.toString();
    }

    public String getConnectionString() {
        if (connectionString == null) {
            StringBuffer result = new StringBuffer();
            result.append(getConnectionURL());
            result.append(";");
            result.append(getInitString());
            connectionString = result.toString();
        }
        return connectionString;
    }

    private String getInitString() {
        if (initString == null) {
            StringBuffer result = new StringBuffer();
            for(File sqlFile : sqlFiles) {
                if(result.length() > 0)
                    result.append("\\;");
                result.append("runscript from '" + getFilename(sqlFile) + "'");
            }
            if(result.length() > 0)
                result.insert(0, "INIT=");
            initString = result.toString();

        }
        return  initString;
    }

    public Connection connect() {
        if (connection == null) {
            try {
                logger.info("Connecting to " + getConnectionString());
                this.connection = DriverManager.getConnection(getConnectionString());
            } catch (SQLException e) {
                throw new RuntimeException("Error connecting DB: " + getConnectionString(),e);
            }
        } else {
            logger.info("Connection to [" + getConnectionURL() + "] is already set.");
        }
        return connection;
    }

    public Connection getConnection() {
        return connection;
    }


    public void close() {

    }

    public void removeTemporaryFiles() {

            if (temporaryConfigFile != null)
                if (!temporaryConfigFile.delete())
                    logger.warn("Could not delete temporary configuration file [" + temporaryConfigFile.getAbsolutePath() + "]");

            if (temporaryWorkingDirectory != null)
                if (!temporaryWorkingDirectory.delete())
                    logger.warn("Could not delete temporary configuration file folder [" + temporaryWorkingDirectory.getAbsolutePath() + "]");

    }

    public String getFilename(File file) {
        return file.getAbsolutePath().replace("\\", "/");
    }

    /**
     * The file with the hibernate configuration takes placed in a (temporary) folder selected by this class.
     * @return
     */
    public File createTemporaryHibernateConfiguration() {
        File configFile = createTemporaryEnvironment();
        createHibernateConfiguration(configFile);
        return configFile;
    }

    /**
     * The file with the hibernate configuration takes placed in a (temporary) folder selected by this class.
     * @return
     */
    public File createTemporaryEnvironment() {
        if (temporaryConfigFile == null) {
            if (temporaryWorkingDirectory == null)
                temporaryWorkingDirectory = Files.createTempDir();
            temporaryConfigFile = createTemporaryFile(temporaryWorkingDirectory);
        }
        return temporaryConfigFile;
    }

    public File createTemporaryFile(File targetDir) {
        File tempFile = null;
        try {
            tempFile = File.createTempFile("hibernate_", ".cfg.xml", targetDir);
        } catch (IOException e) {
            throw new RuntimeException("Error creating temporary file for hibernate configuration.",e);
        }
        return tempFile;
    }

    /**
     * The file with the hibernate configuration takes place in a given file.
     * @param targetFile
     * @return
     */
    public File createHibernateConfiguration(File targetFile) {
        try {
            String content = Resources.toString(HIBERNATE_CFG_TEMPLATE, Charset.defaultCharset());
            content = content.replace("${connection.url}", getConnectionString() );
            content = content.replace("${connection.classes}", getClassList() );
            logger.debug(content);
            Files.write(content, targetFile, Charset.defaultCharset());
            logger.info("The hibernate configuration files placed at " + targetFile.getAbsolutePath());
        } catch (IOException e) {
            String msg = "Error writing file " + targetFile.getAbsolutePath();
            logger.error(msg,e);
            throw new RuntimeException(msg,e);
        }
        return targetFile;
    }

    /**
     * The file with the hibernate configuration takes place in a given file.
     * @param resource
     * @return
     */
    public File createHibernateConfigurationFromResource(URL resource, File targetDir) {
        temporaryWorkingDirectory = targetDir;
        temporaryConfigFile = createTemporaryFile(targetDir);
        try {
            String content = Resources.toString(resource, Charset.defaultCharset());
            Files.write(content, temporaryConfigFile, Charset.defaultCharset());
            logger.info("The hibernate configuration files placed at " + temporaryConfigFile.getAbsolutePath());
        } catch (IOException e) {
            String msg = "Error writing file " + temporaryConfigFile.getAbsolutePath();
            logger.error(msg,e);
            throw new RuntimeException(msg,e);
        }
        return temporaryConfigFile;
    }

    /**
     * To build a connection with a specific database location this methood might be override.
     *
     * @return
     */
    protected String getConnectionURL() {
        return connectionPrefix;
    }

}
